#ifndef SENTENCE_H
#define SENTENCE_H

#include <sys/types.h>
#include <stdio.h>

void sentence(const char *cmd, nl_catd catd, FILE *in, const char *file, void (*process)(const char *, size_t, const char *, int), const char *lang);

#endif
